#include "g95.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#undef G95X_ADD_ADDR
#undef G95X_INDENT
#define G95X_CR ""
#define G95X_BR "<%sL/>"
#define G95X_CRBR "\n" G95X_BR


static char *prefix;

static int pstack = -1;
static FILE *out = NULL;

static char *
f2xml (const char *file)
{
  static char xml_f[256];
  int length;
  int i;

  length = strlen (file);
  strcpy (xml_f, file);

  for (i = length; i >= 0; i--)
    if (xml_f[i] == '.')
      break;
  if (i == 0)
    i = length - 1;


  xml_f[i + 1] = 'x';
  xml_f[i + 2] = 'm';
  xml_f[i + 3] = 'l';
  xml_f[i + 4] = '\0';

  return xml_f;
}

#ifdef G95X_INDENT
static void
indent (int k)
{
  int i;
  for (i = 0; i < pstack + k; i++)
    fprintf (out, "  ");
}
#else
#define indent( k ) do{}while(0)
#endif



int
g95x_init_options (int argc, char *argv[])
{
  out = fopen (f2xml (argv[argc - 1]), "w");
  return 1;
}


g95x_pointer
g95x_get_obj_id_an ()
{
  g95x_bbt_id *st = g95x_new_bbt_id_an (&root_bbt_id);
  return st->q;
}

static g95x_pointer
get_obj_id (int def, g95x_pointer p, int ni)
{
  g95x_bbt_id *st = g95x_get_bbt_id (p, NULL, &root_bbt_id);
  if (!st)
    st = g95x_new_bbt_id (p, NULL, &root_bbt_id);
  if (def > 0)
    {
      st->defined++;
      st->hidden = ni ? 1 : 0;
    }
  else
    {
      st->referenced = 1;
    }
  return st->q;
}

static g95x_pointer
get_als_id (int def, const char *als, g95x_pointer p)
{
  g95x_bbt_id *st = g95x_get_bbt_id (p, als, &root_bbt_id);
  if (!st)
    st = g95x_new_bbt_id (p, als, &root_bbt_id);
  if (def > 0)
    st->defined++;
  return st->q;
}

int
g95x_exists_obj_id (g95x_pointer p)
{
  return g95x_get_bbt_id (p, NULL, &root_bbt_id) != NULL;
}

int
g95x_exists_als_id (const char *als, g95x_pointer p)
{
  return g95x_get_bbt_id (p, als, &root_bbt_id) != NULL;
}

int
g95x_defined_obj_id (g95x_pointer p)
{
  g95x_bbt_id *st;
  if ((st = g95x_get_bbt_id (p, NULL, &root_bbt_id)))
    if (st->defined)
      return 1;
  return 0;
}

int
g95x_defined_als_id (const char *als, g95x_pointer p)
{
  g95x_bbt_id *st;
  if ((st = g95x_get_bbt_id (p, als, &root_bbt_id)))
    if (st->defined)
      return 1;
  return 0;
}


#define MAX_OBJ 1000
static struct
{
  enum
  {
    G95X_STACK_OBJ = 0, G95X_STACK_LST = 1,
    G95X_STACK_SCT = 2, G95X_STACK_FLE = 3,
    G95X_STACK_INC = 4
  } type;
  int nelems;
  const char *name;
  int end;
  g95_file *f;
  int done;
} stack[MAX_OBJ];

static void
o_error ()
{
  fflush (out);
  g95x_die ("");
}




int
g95x_add_dmp ()
{
  if (g95x_option.xmlns)
    {
      int len = strlen (g95x_option.xmlns);
      prefix = (char *) malloc (len + 2);
      strcpy (prefix, g95x_option.xmlns);
      strcat (prefix, ":");
    }
  else
    {
      static char c = '\0';
      prefix = &c;
    }
  if (!g95x_option.xml_no_header)
    fprintf (out, "<?xml version=\"1.0\"?>");
  if (g95x_option.stylesheet)
    fprintf (out, "<?xml-stylesheet type=\"%s\" href=\"%s\"?>",
	     g95x_option.stylesheet_type ? g95x_option.
	     stylesheet_type : "text/css", g95x_option.stylesheet);
  if (g95x_option.xml_xul) 
    fprintf (out, "<window flex=\"1\" xmlns=\"http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul\"><scrollbox flex=\"1\" style=\"overflow:auto;\">");
  fprintf (out, "<%sfortran95 xmlns", prefix);
  if (g95x_option.xmlns)
    fprintf (out, ":%s", g95x_option.xmlns);
  fprintf (out, "=\"%s\">",
	   g95x_option.xmlns_uri ? g95x_option.
	   xmlns_uri : "http://g95-xml.sourceforge.net/");

  return 1;
}

int
g95x_end_dmp ()
{
  fprintf (out, "</fortran95>");
  if (g95x_option.xml_xul) 
    fprintf (out, "</scrollbox></window>");
  fclose (out);
  out = NULL;
  return 1;
}

static int
add_lst1 ()
{
  if (pstack < 0)
    return 0;
  if (stack[pstack].type != G95X_STACK_LST)
    return 0;
  if (stack[pstack].nelems == 0)
    {
      if (pstack > 0)
	{
	  if (stack[pstack - 1].nelems == 0)
	    {
	      if (stack[pstack - 1].end++ == 0)
		fprintf (out, ">" G95X_CR);
	    }
	  stack[pstack - 1].nelems++;
	}
      indent (0);
      if (stack[pstack].done++ == 0)
	return fprintf (out, "<%s%s>" G95X_CR, prefix, stack[pstack].name);
    }

  stack[pstack].nelems++;

  return 0;
}


int
g95x_add_als1 (const char *cls, const char *als, const void *obj)
{
  g95x_pointer valo;
  int an = ! g95x_option.xml_defn;
  

  if (obj == NULL)
    o_error ();

  add_lst1 ();

  valo = get_als_id (1, als, (g95x_pointer) obj);

  stack[++pstack].type = G95X_STACK_OBJ;
  stack[pstack].nelems = 0;
  stack[pstack].name = cls;
  stack[pstack].end = 0;
  stack[pstack].f = NULL;

  if (pstack > 0)
    stack[pstack - 1].nelems++;

  indent (0);
  return fprintf (out, "<%s%s", prefix, cls)
    + ( an == 0 ? fprintf (" id=\"0x%x\"", valo) : 0 )
#ifdef G95X_ADD_ADDR
    + fprintf (out, " addr=\"%s-0x%x\"", als, obj)
#endif
    ;
}

int
g95x_add_obj1 (const char *cls, int ni, const void *obj)
{
  g95x_pointer valo;
  int an = ! g95x_option.xml_defn;

  if (obj == NULL)
    o_error ();

  add_lst1 ();

  if (an == 0) 
  {
    valo = get_obj_id (1, (g95x_pointer) obj, ni);
    if (ni == 0)
      g95x_die ("");
  }

  stack[++pstack].type = G95X_STACK_OBJ;
  stack[pstack].nelems = 0;
  stack[pstack].name = cls;
  stack[pstack].end = 0;
  stack[pstack].f = NULL;

  if (pstack > 0)
    stack[pstack - 1].nelems++;

  indent (0);

  if ((an == 0) && (ni == 0))
    return fprintf (out, "<%s%s id=\"0x%x\"", prefix, cls, valo)
#ifdef G95X_ADD_ADDR
      + fprintf (out, " addr=\"0x%x\"", obj)
#endif
      ;
  else
    return fprintf (out, "<%s%s", prefix, cls);

  return 0;
}


int
g95x_end_obj1 ()
{
  if (stack[pstack].type != G95X_STACK_OBJ)
    o_error ();
  pstack--;
  if (pstack < -1)
    o_error ();

  if (stack[pstack + 1].nelems > 0)
    {
      indent (1);
      return fprintf (out, "</%s%s>" G95X_CR, prefix, stack[pstack + 1].name);
    }
  else
    {
      return fprintf (out, "/>" G95X_CR);
    }
  return 0;
}

int
g95x_add_inc ()
{
  add_lst1 ();

  stack[++pstack].type = G95X_STACK_INC;
  stack[pstack].name = "include";
  stack[pstack].end = 0;

  indent (0);

  return fprintf (out, G95X_CRBR "<%sinclude>", prefix, prefix);
}

int
g95x_end_inc ()
{
  if (stack[pstack].type != G95X_STACK_INC)
    o_error ();
  pstack--;
  if (pstack < -1)
    o_error ();

  indent (1);
  fprintf (out, "</%s%s>", prefix, stack[pstack + 1].name);

  return 0;
}

int
g95x_add_fle (g95_file * f)
{
  g95x_pointer valo;


  valo = get_obj_id (1, (g95x_pointer) f, 0);

  stack[++pstack].type = G95X_STACK_FLE;
  stack[pstack].name = "file";
  stack[pstack].end = 0;
  stack[pstack].f = f;

  indent (0);

  return fprintf (out, "<%sfile", prefix)
#ifdef G95X_ADD_ADDR
    + fprintf (out, " addr=\"0x%x\"", obj)
#endif
    + fprintf (out, " name=\"%s\" width=\"%d\" form=\"%s\">" G95X_BR,
	       f->filename, g95x_get_line_width (),
	       g95x_form_type (f->x.form), prefix);
  ;


  return 0;
}

int
g95x_end_fle ()
{
  if (stack[pstack].type != G95X_STACK_FLE)
    o_error ();
  pstack--;
  if (pstack < -1)
    o_error ();

  indent (1);
  fprintf (out, "</%s%s>" G95X_CR, prefix, stack[pstack + 1].name);

  return 0;
}


int
g95x_add_att1_cst (const char *key, const char *vals)
{
  const char *s;

  fprintf (out, " %s=\"", key);

  for (s = vals; *s != 0; s++)
    {
      switch (*s)
	{
	case '"':
	  fprintf (out, "&quot;");
	  break;
	case '<':
	  fprintf (out, "&lt;");
	  break;
	case '>':
	  fprintf (out, "&gt;");
	  break;
	case '&':
	  fprintf (out, "&amp;");
	  break;
	case '\\':
	  fprintf (out, "\\\\");
	  break;
	default:
	  if (isprint (*s))
	    fprintf (out, "%c", *s);
	  else
	    fprintf (out, "&#%d;", (unsigned char) *s);
	}
    }

  fprintf (out, "\"");

  return 1;
}

int
g95x_add_att1_str (const char *key, const char *vals)
{
  return fprintf (out, " %s=\"%s\"", key, vals);
}

int
g95x_add_att1_int (const char *key, const int vali)
{
  if ((!strcmp (key, "c1") || (!strcmp (key, "c2"))))
    {
      if (vali == 0)
	g95x_die ("");
    }
  return fprintf (out, " %s=\"%d\"", key, vali);
}

int
g95x_add_att1_obj (const char *key, const void *valo)
{
  return fprintf (out, " %s=\"0x%x\"", key,
		  get_obj_id (0, (g95x_pointer) valo, 0))
#ifdef G95X_ADD_ADDR
    + fprintf (out, " addr-%s=\"0x%x\"", key, valo)
#endif
    ;
}

int
g95x_add_att1_als (const char *key, const char *als, const void *valo)
{
  return fprintf (out, " %s=\"0x%x\"", key,
		  get_als_id (0, als, (g95x_pointer) valo))
#ifdef G95X_ADD_ADDR
    + fprintf (out, " %s-addr=\"%s-0x%x\"", key, als, valo)
#endif
    ;
}

int
g95x_add_att1_und (const char *key)
{
  return fprintf (out, " %s=\"undef\"", key);
}

int
g95x_add_att1_chr (const char *key, char valc)
{
  return fprintf (out, " %s=\"%c\"", key, valc);
}

static int
end_att1 ()
{
  if (pstack < 0)
    o_error ();
  if (stack[pstack].type != G95X_STACK_OBJ)
    o_error ();
  if (stack[pstack].end++)
    return 0;
  return fprintf (out, ">" G95X_CR);
}

int
g95x_end_att1 ()
{
  return 0;
}



int
g95x_add_lst1 (const char *lst)
{

  stack[++pstack].type = G95X_STACK_LST;
  stack[pstack].nelems = 0;
  stack[pstack].name = lst;
  stack[pstack].end = 0;
  stack[pstack].f = NULL;
  stack[pstack].done = 0;

  return 1;
}


int
g95x_end_lst1 ()
{
  int p = 0;

  while (stack[pstack].type == G95X_STACK_FLE)
    {
      g95x_close_file (stack[pstack].f);
      p++;
    }

  if ((stack[pstack].nelems > 0) || (p > 0))
    indent (0);

  if (stack[pstack].type != G95X_STACK_LST)
    o_error ();
  pstack--;
  if (pstack < -1)
    o_error ();

  if ((stack[pstack + 1].nelems > 0) || (p > 0))
    return fprintf (out, "</%s%s>" G95X_CR, prefix, stack[pstack + 1].name);
  return 0;
}


int
g95x_psh_lst1_obj (const void *valo)
{
  add_lst1 ();
  indent (1);

  stack[pstack].nelems++;

  return fprintf (out, "<%sref val=\"0x%x\"", prefix,
		  get_obj_id (0, (g95x_pointer) valo, 0))
#ifdef G95X_ADD_ADDR
    + fprintf (out, " addr=\"0x%x\"", valo)
#endif
    + fprintf (out, "/>" G95X_CR);
}

int
g95x_psh_lst1_als (const char *als, const void *valo)
{
  add_lst1 ();
  indent (1);

  stack[pstack].nelems++;

  return fprintf (out, "<%sref val=\"0x%x\"/>" G95X_CR, prefix,
		  get_als_id (0, als, (g95x_pointer) valo));
}

int
g95x_psh_lst1_chr (char valc)
{
  add_lst1 ();
  indent (1);

  stack[pstack].nelems++;

  return fprintf (out, "<%sc val=\"%c\"/>" G95X_CR, prefix, valc);
}

int
g95x_psh_lst1_und ()
{
  add_lst1 ();
  indent (1);

  stack[pstack].nelems++;

  return fprintf (out, "<%sundef/>" G95X_CR, prefix);
}

int
g95x_psh_lst1_str (const char *vals)
{
  add_lst1 ();
  indent (1);

  stack[pstack].nelems++;

  return fprintf (out, "<%sstr val=\"%s\"/>" G95X_CR, prefix, vals);
}

int
g95x_psh_lst1_int (int vali)
{
  add_lst1 ();
  indent (1);

  stack[pstack].nelems++;

  return fprintf (out, "<%sint val=\"%d\"/>" G95X_CR, prefix, vali);
}

int
g95x_add_sct1 (const char *sct)
{
  if (sct == NULL)
    o_error ();

  if (stack[pstack].type != G95X_STACK_INC)
    end_att1 ();

  stack[++pstack].type = G95X_STACK_SCT;
  stack[pstack].nelems = 0;
  stack[pstack].name = sct;
  stack[pstack].end = 0;
  stack[pstack].f = NULL;

  if (pstack > 0)
    stack[pstack - 1].nelems++;

  indent (0);
  return fprintf (out, "<%s%s>" G95X_CR, prefix, sct);
}

int
g95x_end_sct1 ()
{
  indent (0);
  if (stack[pstack].type != G95X_STACK_SCT)
    o_error ();
  pstack--;
  return fprintf (out, "</%s%s>" G95X_CR, prefix, stack[pstack + 1].name);
}


int
g95x_add_txt1 ()
{
  int txt_indent = 0;

  switch (stack[pstack].type)
    {
    case G95X_STACK_OBJ:
      end_att1 ();
      txt_indent = 1;
      break;
    case G95X_STACK_LST:
      if (stack[pstack].nelems > 0)
	{
	  txt_indent = 1;
	}
      else
	{
	  pstack--;
	  end_att1 ();
	  pstack++;
	  txt_indent = 0;
	}
      break;
    case G95X_STACK_FLE:
    case G95X_STACK_INC:
      txt_indent = 1;
      break;
    default:
      txt_indent = 0;
    }
  indent (txt_indent);
  return 0;
}

int g95x_put_txt1_cd = 0;

int
g95x_put_txt1 (const char *key, const char *txt, int cpp, int id, int *open)
{
  if (*open == 0)
    {
      *open = 1;
      g95x_add_txt1 ();
    }

  if (!strcmp ("code", key)) 
  {
    if (!g95x_option.xml_sc)
      if ((g95x_put_txt1_cd == 0) && (cpp == 0) && txt && txt[0])
      {
        return fprintf (out, "%s", txt);
      }
    key = "c";
  } 
  else if (!strcmp ("comment", key))
  {
    key = "C";
  }
  else if (!strcmp ("br", key))
  {
    return fprintf (out, G95X_CRBR, prefix);
  }

  if (id == 0)
    {
      if ((!strcmp ("space", key)) && (!cpp) && txt && txt[0])
	return fprintf (out, "%s", txt);
      return
	(txt && txt[0]) ?
	cpp ?
	fprintf (out, "<%s%s-cpp>%s</%s%s-cpp>", prefix, key, txt, prefix,
		 key) : fprintf (out, "<%s%s>%s</%s%s>", prefix, key, txt,
				 prefix, key) : fprintf (out, "<%s%s/>",
							 prefix, key);
    }
  else
    {
      return
	(txt && txt[0]) ?
	cpp ?
	fprintf (out, "<%s%s-cpp cpp=\"0x%x\">%s</%s%s-cpp>", prefix, key, id,
		 txt, prefix, key) : fprintf (out,
					      "<%s%s id=\"0x%x\">%s</%s%s>",
					      prefix, key, id, txt, prefix,
					      key) : fprintf (out,
							      "<%s%s id=\"0x%x\"/>",
							      prefix, key,
							      id);
    }
  return 0;
}

int
g95x_end_txt1 (int *open)
{
  if (*open == 0)
    return 0;
  fprintf (out, G95X_CR);
  if (stack[pstack].type == G95X_STACK_LST)
    if (stack[pstack].nelems == 0)
      add_lst1 ();
  stack[pstack].nelems++;
  *open = 0;
  return 0;
}
